/*
 * EvidenceMarkerGeneratorPlugin.java 20 Nov 2018
 *
 * Copyright (c) 2018 Emmanuel PUYBARET / eTeks <info@eteks.com>. All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package com.eteks.sweethome3d.plugin.markergenerator;

import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.RenderingHints;
import java.awt.font.TextLayout;
import java.awt.geom.Area;
import java.awt.geom.Rectangle2D;
import java.awt.image.BufferedImage;
import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.URL;
import java.net.URLConnection;
import java.net.URLStreamHandler;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Locale;
import java.util.ResourceBundle;

import javax.imageio.ImageIO;
import javax.jnlp.BasicService;
import javax.jnlp.ServiceManager;
import javax.jnlp.UnavailableServiceException;
import javax.media.j3d.BranchGroup;
import javax.media.j3d.Transform3D;
import javax.swing.JComponent;
import javax.swing.JEditorPane;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JSpinner;
import javax.swing.KeyStroke;
import javax.swing.event.HyperlinkEvent;
import javax.swing.event.HyperlinkListener;
import javax.swing.undo.AbstractUndoableEdit;
import javax.swing.undo.UndoableEditSupport;
import javax.vecmath.Point3f;
import javax.vecmath.Vector3f;

import com.eteks.sweethome3d.j3d.ModelManager;
import com.eteks.sweethome3d.j3d.Object3DBranch;
import com.eteks.sweethome3d.model.CatalogPieceOfFurniture;
import com.eteks.sweethome3d.model.HomePieceOfFurniture;
import com.eteks.sweethome3d.model.Selectable;
import com.eteks.sweethome3d.model.UserPreferences;
import com.eteks.sweethome3d.plugin.Plugin;
import com.eteks.sweethome3d.plugin.PluginAction;
import com.eteks.sweethome3d.swing.NullableSpinner;
import com.eteks.sweethome3d.swing.NullableSpinner.NullableSpinnerNumberModel;
import com.eteks.sweethome3d.swing.SwingTools;
import com.eteks.sweethome3d.tools.OperatingSystem;
import com.eteks.sweethome3d.tools.ResourceURLContent;
import com.eteks.sweethome3d.viewcontroller.HomeView;

/**
 * A plug-in that let the user creates numbered evidence markers.
 * @author Emmanuel Puybaret
 */
public class EvidenceMarkerGeneratorPlugin extends Plugin {
  private static final String EVIDENCE_MARKER_ID = "eTeks#evidenceMarker";

  private static String markerModelBase = null;

  @Override
  public PluginAction [] getActions() {
    return new PluginAction [] {new PluginAction(
            "com.eteks.sweethome3d.plugin.markergenerator.ApplicationPlugin",
            "ADD_EVIDENCE_MARKER", getPluginClassLoader(), true) {
        @Override
        public void execute() {
          addEvidenceMarker();
        }
      }
    };
  }

  /**
   * Creates an evidence marker from the number entered by the user.
   */
  private void addEvidenceMarker() {
    final ResourceBundle resource = ResourceBundle.getBundle(
        "com.eteks.sweethome3d.plugin.markergenerator.ApplicationPlugin",
        Locale.getDefault(), getPluginClassLoader());
    final HomeView homeView = getHomeController().getView();

    try {
      // Ignore plug-in in protected Java Web Start environment
      ServiceManager.lookup("javax.jnlp.FileSaveService");
      // Use an uneditable editor pane to let user select text in dialog
      JEditorPane messagePane = new JEditorPane("text/html",
          resource.getString("shapeGeneratorJavaWebStartInfo.message"));
      messagePane.setOpaque(false);
      messagePane.setEditable(false);
      try {
        // Lookup the javax.jnlp.BasicService object
        final BasicService service = (BasicService)ServiceManager.lookup("javax.jnlp.BasicService");
        // If basic service supports  web browser
        if (service.isWebBrowserSupported()) {
          // Add a listener that displays hyperlinks content in browser
          messagePane.addHyperlinkListener(new HyperlinkListener() {
            public void hyperlinkUpdate(HyperlinkEvent ev) {
              if (ev.getEventType() == HyperlinkEvent.EventType.ACTIVATED) {
                service.showDocument(ev.getURL());
              }
            }
          });
        }
      } catch (UnavailableServiceException ex) {
        // Too bad : service is unavailable
      }

      String title = resource.getString("shapeGeneratorJavaWebStartInfo.title");
      JOptionPane.showMessageDialog((JComponent)homeView, messagePane, title, JOptionPane.WARNING_MESSAGE);
      return;
    } catch (UnavailableServiceException ex) {
    }

    // Create components
    JPanel markerPanel = new JPanel(new GridBagLayout());
    UserPreferences preferences = getUserPreferences();
    int labelAlignment = OperatingSystem.isMacOSX()
        ? GridBagConstraints.LINE_END
        : GridBagConstraints.LINE_START;

    JLabel numberLabel = new JLabel(resource.getString("numberLabel.text"));
    markerPanel.add(numberLabel,
        new GridBagConstraints(0, 0, 1, 1, 0, 0, labelAlignment,
            GridBagConstraints.NONE, new Insets(0, 0, 5, 5), 0, 0));
    int max = 1;
    for (HomePieceOfFurniture piece : getHome().getFurniture()) {
      if (EVIDENCE_MARKER_ID.equals(piece.getCatalogId())) {
        max++;
      }
    }
    NullableSpinnerNumberModel numberSpinnerModel = new NullableSpinner.NullableSpinnerNumberModel(max, 1, 99, 1);
    NullableSpinner numberSpinner = new NullableSpinner(numberSpinnerModel);
    markerPanel.add(numberSpinner,
        new GridBagConstraints(1, 0, 1, 1, 0, 0, GridBagConstraints.LINE_START,
            GridBagConstraints.HORIZONTAL, new Insets(0, 0, 5, 0), 0, 0));

    Rectangle2D selectionRectangle = null;
    float defaultElevation = 0;
    for (Selectable item : getHome().getSelectedItems()) {
      for (float [] point : item.getPoints()) {
        if (selectionRectangle == null) {
          selectionRectangle = new Rectangle2D.Float(point [0], point [1], 0, 0);
        } else {
          selectionRectangle.add(point [0], point [1]);
        }
      }
    }
    for (Selectable item : getHome().getSelectedItems()) {
      if (item instanceof HomePieceOfFurniture
          && item.containsPoint((float)selectionRectangle.getCenterX(), (float)selectionRectangle.getCenterY(), 0)) {
        HomePieceOfFurniture piece = (HomePieceOfFurniture)item;
        defaultElevation = Math.max(defaultElevation, piece.getDropOnTopElevation() * piece.getHeight() + piece.getElevation());
      }
    }
    if (selectionRectangle == null) {
      selectionRectangle = new Rectangle2D.Float(0, 0, 0, 0);
    }

    JLabel xLabel = new JLabel(String.format(resource.getString("xLabel.text"), preferences.getLengthUnit().getName()));
    markerPanel.add(xLabel,
        new GridBagConstraints(0, 1, 1, 1, 0, 0, labelAlignment,
            GridBagConstraints.NONE, new Insets(0, 0, 5, 5), 0, 0));
    NullableSpinnerNumberModel xSpinnerModel = new NullableSpinner.NullableSpinnerLengthModel(preferences, (float)selectionRectangle.getCenterX(), -100000f, 100000f);
    NullableSpinner xSpinner = new NullableSpinner(xSpinnerModel);
    markerPanel.add(xSpinner,
        new GridBagConstraints(1, 1, 1, 1, 0, 0, GridBagConstraints.LINE_START,
            GridBagConstraints.HORIZONTAL, new Insets(0, 0, 5, 0), 0, 0));

    JLabel yLabel = new JLabel(String.format(resource.getString("yLabel.text"), preferences.getLengthUnit().getName()));
    markerPanel.add(yLabel,
        new GridBagConstraints(0, 2, 1, 1, 0, 0, labelAlignment,
            GridBagConstraints.NONE, new Insets(0, 0, 5, 5), 0, 0));
    NullableSpinnerNumberModel ySpinnerModel = new NullableSpinner.NullableSpinnerLengthModel(preferences, (float)selectionRectangle.getCenterY(), -100000f, 100000f);
    NullableSpinner ySpinner = new NullableSpinner(ySpinnerModel);
    markerPanel.add(ySpinner,
        new GridBagConstraints(1, 2, 1, 1, 0, 0, GridBagConstraints.LINE_START,
            GridBagConstraints.HORIZONTAL, new Insets(0, 0, 5, 0), 0, 0));

    JLabel elevationLabel = new JLabel(String.format(resource.getString("elevationLabel.text"), preferences.getLengthUnit().getName()));
    markerPanel.add(elevationLabel,
        new GridBagConstraints(0, 3, 1, 1, 0, 0, labelAlignment,
            GridBagConstraints.NONE, new Insets(0, 0, 0, 5), 0, 0));
    NullableSpinnerNumberModel elevationSpinnerModel = new NullableSpinner.NullableSpinnerLengthModel(preferences, defaultElevation, 0, 10000f);
    NullableSpinner elevationSpinner = new NullableSpinner(elevationSpinnerModel);
    markerPanel.add(elevationSpinner,
        new GridBagConstraints(1, 3, 1, 1, 0, 0, GridBagConstraints.LINE_START,
            GridBagConstraints.HORIZONTAL, new Insets(0, 0, 0, 0), 0, 0));

    if (!OperatingSystem.isMacOSX()) {
      numberLabel.setDisplayedMnemonic(KeyStroke.getKeyStroke(resource.getString("numberLabel.mnemonic")).getKeyCode());
      numberLabel.setLabelFor(numberSpinner);
      xLabel.setDisplayedMnemonic(KeyStroke.getKeyStroke(resource.getString("xLabel.mnemonic")).getKeyCode());
      xLabel.setLabelFor(xSpinner);
      yLabel.setDisplayedMnemonic(KeyStroke.getKeyStroke(resource.getString("yLabel.mnemonic")).getKeyCode());
      yLabel.setLabelFor(ySpinner);
      elevationLabel.setDisplayedMnemonic(KeyStroke.getKeyStroke(resource.getString("elevationLabel.mnemonic")).getKeyCode());
      elevationLabel.setLabelFor(elevationSpinner);
    }

    if (SwingTools.showConfirmDialog((JComponent)homeView, markerPanel,
        resource.getString("ADD_EVIDENCE_MARKER.NAME"), ((JSpinner.DefaultEditor)numberSpinner.getEditor()).getTextField()) == JOptionPane.OK_OPTION) {
      final int number = ((Number)numberSpinnerModel.getValue()).intValue();
      final float x = ((Number)xSpinnerModel.getValue()).floatValue();
      final float y = ((Number)ySpinnerModel.getValue()).floatValue();
      final float elevation = ((Number)elevationSpinnerModel.getValue()).floatValue();

      if (markerModelBase == null) {
        // Read evidence marker base model
        final StringBuilder model = new StringBuilder();
        BufferedReader reader = null;
        try {
          reader = new BufferedReader(new InputStreamReader(
              EvidenceMarkerGeneratorPlugin.class.getResource("evidenceMarker.obj").openStream(), "ISO-8859-1"));
          for (String s; (s = reader.readLine()) != null; ) {
            model.append(s);
            model.append("\n");
          }
        } catch (IOException ex) {
          // Shouldn't happen
          ex.printStackTrace();
        } finally {
          if (reader != null) {
            try {
              reader.close();
            } catch (IOException ex) {
            }
          }
        }

        markerModelBase = model.toString();
      }

      // Create plan icon
      BufferedImage planIconImage = new BufferedImage(128, 100, BufferedImage.TYPE_INT_ARGB);
      Graphics2D g2D = (Graphics2D)planIconImage.getGraphics();
      g2D.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
      g2D.setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING, RenderingHints.VALUE_TEXT_ANTIALIAS_ON);
      g2D.setRenderingHint(RenderingHints.KEY_FRACTIONALMETRICS, RenderingHints.VALUE_FRACTIONALMETRICS_ON);
      Font font = new Font("Helvetica", Font.BOLD, 7);
      String text = String.valueOf(number);
      Rectangle2D size = font.getStringBounds(text, g2D.getFontRenderContext());
      g2D.setColor(new Color(247, 204, 69));
      g2D.fillRect(0, 0, planIconImage.getWidth() - 1, planIconImage.getHeight() - 1);
      g2D.setColor(Color.BLACK);
      if (number == 6 || number == 9|| number == 66 || number == 69 || number == 99) {
        g2D.fillRect(planIconImage.getWidth() / 2 - 10, planIconImage.getHeight() - 12, 19, 4);
      }
      g2D.translate((planIconImage.getWidth() - size.getWidth() * 13) / 2, planIconImage.getHeight() - 19);
      g2D.scale(13, 13);
      g2D.setFont(font);
      g2D.drawString(text, 0, 0);
      ByteArrayOutputStream planIconOutput = new ByteArrayOutputStream();
      try {
        ImageIO.write(planIconImage, "PNG", planIconOutput);
      } catch (IOException ex) {
        ex.printStackTrace();
      }

      // Retrieve text points
      final TextLayout textLayout = new TextLayout(text, font, g2D.getFontRenderContext());
      final List<float [][]> textPoints = new ArrayList<float[][]>();
      // Create a dummy subclass of Object3DBranch to call getAreaPoints
      new Object3DBranch() {
          {
            textPoints.addAll(getAreaPoints(new Area(textLayout.getOutline(null)), 0.1f, false));
          }

          public void update() {
          }
        };
      g2D.dispose();

      final StringBuilder markerModel = new StringBuilder(markerModelBase);
      // Generate number shape in 3D model
      markerModel.append("g number\nusemtl flblack\n");
      int i = 25; // Number of vertices in evidenceMarker.obj
      Transform3D transform = new Transform3D();
      transform.rotX(1.12);
      Transform3D translation = new Transform3D();
      transform.setTranslation(new Vector3f(-(float)size.getWidth() / 2, -2f, 2.9f));
      transform.mul(translation, transform);
      for (float [][] points : textPoints) {
        for (float [] point : points) {
          Point3f p = new Point3f(point [0], 0, point [1]);
          transform.transform(p);
          markerModel.append("v " + p.x + " " + p.y + " " + p.z + "\n");
        }
        markerModel.append("f");
        for (int j = 0; j < points.length; j++) {
          markerModel.append(" " + i++);
        }
        markerModel.append("\n");
      }
      // Opposite side
      Transform3D rotation = new Transform3D();
      rotation.rotY(Math.PI);
      transform.mul(rotation, transform);
      for (float [][] points : textPoints) {
        for (float [] point : points) {
          Point3f p = new Point3f(point [0], 0, point [1]);
          transform.transform(p);
          markerModel.append("v " + p.x + " " + p.y + " " + p.z + "\n");
        }
        markerModel.append("f");
        for (int j = 0; j < points.length; j++) {
          markerModel.append(" " + i++);
        }
        markerModel.append("\n");
      }

      // Create catalog icon
      BufferedImage iconImage = new BufferedImage(32, 32, BufferedImage.TYPE_INT_ARGB);
      g2D = (Graphics2D)iconImage.getGraphics();
      g2D.setRenderingHint(RenderingHints.KEY_RENDERING, RenderingHints.VALUE_RENDER_QUALITY);
      g2D.scale(0.25, 0.25);
      g2D.drawImage(planIconImage, null, 0, 14);
      g2D.dispose();
      ByteArrayOutputStream iconOutput = new ByteArrayOutputStream();
      try {
        ImageIO.write(iconImage, "PNG", iconOutput);
      } catch (IOException ex) {
        ex.printStackTrace();
      }

      ResourceURLContent iconContent = null;
      ResourceURLContent planIconContent = null;
      ResourceURLContent modelContent = null;
      try {
        // Access to icons and 3D model from memory
        iconContent = new ResourceURLContent(new URL(null, "memory://evidenceMarkerIcon_" + number,
            new MemoryStreamHandler(iconOutput.toByteArray())), false);
        planIconContent = new ResourceURLContent(new URL(null, "memory://evidenceMarkerPlanIcon_" + number,
            new MemoryStreamHandler(planIconOutput.toByteArray())), false);
        modelContent = new ResourceURLContent(new URL(null, "memory://evidenceMarkerModel_" + number,
            new MemoryStreamHandler(markerModel.toString().getBytes("ISO-8859-1"))), false);
      } catch (IOException ex) {
        // Can't happen
      }

      final ResourceURLContent evidenceMarkerIconContent = iconContent;
      final ResourceURLContent evidenceMarkerPlanIconContent = planIconContent;
      final ResourceURLContent evidenceMarkerModelContent = modelContent;
      ModelManager.getInstance().loadModel(evidenceMarkerModelContent, new ModelManager.ModelObserver() {
          public void modelUpdated(BranchGroup modelRoot) {
            Vector3f size = ModelManager.getInstance().getSize(modelRoot);
            HomePieceOfFurniture piece = new HomePieceOfFurniture(new CatalogPieceOfFurniture(
                  EVIDENCE_MARKER_ID, String.format(resource.getString("evidenceMarkerName"), number), null,
                  evidenceMarkerIconContent, evidenceMarkerPlanIconContent, evidenceMarkerModelContent,
                  size.x, size.z, size.y,
                  elevation, true, null, System.getProperty("user.name"), true, null, null));
            piece.setX(x);
            piece.setY(y);

            UndoableEditSupport undoSupport = getUndoableEditSupport();
            undoSupport.beginUpdate();
            getHomeController().getFurnitureController().addFurniture(Arrays.asList(new HomePieceOfFurniture [] {piece}));
            undoSupport.postEdit(new AbstractUndoableEdit() {
                @Override
                public String getPresentationName() {
                  return resource.getString("undoAddEvidenceMarker");
                }
              });

            // End compound edit
            undoSupport.endUpdate();
          }

          public void modelError(Exception ex) {
            // Shouldn't happen since we import a model we just exported
          }
        });
    }
  }

  /**
   * A stream handler able to read bytes in memory.
   */
  private static class MemoryStreamHandler extends URLStreamHandler {
    private final byte [] bytes;

    private MemoryStreamHandler(byte [] bytes) {
      this.bytes = bytes;
    }

    @Override
    protected URLConnection openConnection(URL url) throws IOException {
      return new URLConnection(url) {
          @Override
          public void connect() throws IOException {
            // Don't need to connect
          }

          @Override
          public InputStream getInputStream() throws IOException {
            return new ByteArrayInputStream(bytes);
          }
        };
    }
  }
}
